"""Test suite for the PacketFilter class"""

import os
import time
import unittest
import subprocess
from socket import *

import pf


class TestPacketFilter(unittest.TestCase):
    """Test case class for pf.PacketFilter"""

    testif = "lo0"

    def setUp(self):
        self.pf = pf.PacketFilter()
        self._init_state = {"enabled": self.pf.get_status().running,
                            "ruleset": self.pf.get_ruleset()}
        if not self._init_state["enabled"]:
            self.pf.enable()
        self.pf.clear_rules()

    def tearDown(self):
        self.pf.clear_rules()
        if not self._init_state["enabled"]:
            self.pf.disable()
        else:
            self.pf.load_ruleset(self._init_state["ruleset"])

    def test_enable(self):
        self.pf.disable()
        self.assertFalse(self.pf.get_status().running)
        self.pf.enable()
        self.assertTrue(self.pf.get_status().running)

    def test_set_debug(self):
        _dbg = self.pf.get_status().debug
        for dbg in (pf.LOG_DEBUG, pf.LOG_ERR, _dbg):
            self.pf.set_debug(dbg)
            self.assertEqual(self.pf.get_status().debug , dbg)

    def test_set_hostid(self):
        _id = self.pf.get_status().hostid
        for id in (1234, _id):
            self.pf.set_hostid(id)
            self.assertEqual(self.pf.get_status().hostid , id)

    def test_set_reassembly(self):
        _flags = self.pf.get_status().reass
        flags = pf.PF_REASS_ENABLED | pf.PF_REASS_NODF
        self.pf.set_reassembly(flags)
        self.assertEqual(self.pf.get_status().reass, flags)
        self.pf.set_reassembly(_flags)

    def test_set_limit(self):
        limits = {"states": 5000, "tables": 2000}
        for limit, value in limits.items():
            _value = self.pf.get_limit(limit)
            self.pf.set_limit(limit, value)
            self.assertEqual(self.pf.get_limit(limit), value)
            self.pf.set_limit(limit, _value)

    def test_set_timeout(self):
        timeouts = {"frag": 25, "interval": 20}
        for tmout, value in timeouts.items():
            _value = self.pf.get_timeout(tmout)
            self.pf.set_timeout(tmout, value)
            self.assertEqual(self.pf.get_timeout(tmout), value)
            self.pf.set_timeout(tmout, _value)

    def test_set_ifflags(self):
        _flags = self.pf.get_ifaces(self.testif).flags
        self.pf.clear_ifflags(self.testif)
        self.pf.set_ifflags(self.testif, pf.PFI_IFLAG_SKIP)
        self.assertEqual(self.pf.get_ifaces(self.testif).flags,
                         pf.PFI_IFLAG_SKIP)
        self.pf.set_ifflags(self.testif, _flags)

    def test_set_status_if(self):
        _ifname = self.pf.get_status().ifname
        self.pf.set_status_if(self.testif)
        self.assertEqual(self.pf.get_status().ifname, self.testif)
        self.pf.set_status_if(_ifname)

    def test_clear_status(self):
        self.pf.clear_status()
        self.assertGreaterEqual(pf._utils.uptime(), self.pf.get_status().since)

    def __test_clear_states(self):
        self.pf.clear_rules()
        self._create_state()
        self.assertIsNotNone(filter(lambda s: s.proto == IPPROTO_UDP,
                                    self.pf.get_states()))
        self.pf.clear_states()
        for state in self.pf.get_states():
            self.assertNotEqual(state.proto, IPPROTO_UDP)

    def __test_kill_states(self):
        self.pf.clear_rules()
        self._create_state()
        self.assertEqual(self.pf.kill_states(proto=IPPROTO_UDP), 1)

    def test_clear_rules(self):
        ruleset = pf.PFRuleset()
        ruleset.append(pf.PFRule(action=pf.PF_PASS,
                                 flags="S", flagset="SA",
                                 keep_state=pf.PF_STATE_NORMAL))
        self.pf.load_ruleset(ruleset)
        self.pf.clear_rules()
        self.assertFalse(self.pf.get_ruleset().rules)

    def test_load_queues(self):
        # Rules to load in pf.conf format:
        #   queue std on em0 bandwidth 100M
        #   queue ssh parent std bandwidth 10M burst 90M for 100ms
        #   queue mail parent std bandwidth 10M, min 5M, max 25M
        #   queue http parent std bandwidth 80M default
        ifname  = self.testif
        parentq = "root_" + ifname
        MB = 10**6
        queues = [pf.PFQueue(qname=parentq, ifname=ifname,
                             flags=pf.PFQS_ROOTCLASS,
                             linkshare=pf.ServiceCurve(bandwidth=100*MB)),
                  pf.PFQueue(qname="std", parent=parentq, ifname=ifname,
                             linkshare=pf.ServiceCurve(bandwidth=100*MB)),
                  pf.PFQueue(qname="ssh", parent="std", ifname=ifname,
                             linkshare=pf.ServiceCurve(10*MB, 90*MB, 100)),
                  pf.PFQueue(qname="mail", parent="std", ifname=ifname,
                             linkshare=pf.ServiceCurve(bandwidth=10*MB),
                             realtime=pf.ServiceCurve(bandwidth=5*MB),
                             upperlimit=pf.ServiceCurve(bandwidth=25*MB)),
                  pf.PFQueue(qname="http", parent="std", ifname=ifname,
                             linkshare=pf.ServiceCurve(bandwidth=80*MB),
                             flags=pf.PFQS_DEFAULT)]
        self.pf.clear_rules()
        self.pf.load_queues(*queues)
        self.assertEqual(len(self.pf.get_queues()), len(queues))
        for queue in self.pf.get_queues():
            self.assertIn(queue.qname, map(lambda q: q.qname, queues))

    def test_load_ruleset(self):
        iface = pf.PFAddr(type=pf.PF_ADDR_DYNIFTL, ifname=self.testif)
        tables = [pf.PFTable("web_srv", "10.0.1.20", "10.0.1.21", "10.0.1.22")]
        rules = [
            # match out on $ifname inet from !($ifname) to any nat-to ($ifname)
            pf.PFRule(action=pf.PF_MATCH,
                      direction=pf.PF_OUT,
                      ifname=self.testif,
                      af=AF_INET,
                      src=pf.PFRuleAddr(iface, neg=True),
                      nat=pf.PFPool(pf.PF_POOL_NAT, iface)),
            # pass out quick
            pf.PFRule(action=pf.PF_PASS,
                      direction=pf.PF_OUT,
                      quick=True,
                      flags="S", flagset="SA",
                      keep_state=pf.PF_STATE_NORMAL),
            # anchor "test_anchor"
            pf.PFRuleset("test_anchor"),
            # pass in on $ifname inet proto tcp from any to $ifname port ssh
            pf.PFRule(action=pf.PF_PASS,
                      direction=pf.PF_IN,
                      ifname=self.testif,
                      af=AF_INET,
                      proto=IPPROTO_TCP,
                      dst=pf.PFRuleAddr(iface, pf.PFPort("ssh", IPPROTO_TCP)),
                      flags="S", flagset="SA",
                      keep_state=pf.PF_STATE_NORMAL),
            # pass in on $ifname inet proto tcp to $ifname port www \
            #     rdr-to <web_srv> round-robin sticky-address
            pf.PFRule(action=pf.PF_PASS,
                      direction=pf.PF_IN,
                      ifname=self.testif,
                      af=AF_INET,
                      proto=IPPROTO_TCP,
                      dst=pf.PFRuleAddr(iface, pf.PFPort("www", IPPROTO_TCP)),
                      flags="S", flagset="SA",
                      keep_state=pf.PF_STATE_NORMAL,
                      rdr=pf.PFPool(pf.PF_POOL_RDR, pf.PFAddr("<web_srv>"),
                                    opts=(pf.PF_POOL_ROUNDROBIN|
                                          pf.PF_POOL_STICKYADDR))),
            # pass out on $ifname inet proto tcp to port 80 \
            #     divert-packet port 700
            pf.PFRule(action=pf.PF_PASS,
                      direction=pf.PF_OUT,
                      ifname=self.testif,
                      af=AF_INET,
                      proto=IPPROTO_TCP,
                      dst=pf.PFRuleAddr(port=pf.PFPort("www", IPPROTO_TCP)),
                      divert=pf.PFDivert(pf.PF_DIVERT_PACKET, port=700)),
            # pass in inet proto icmp all icmp-type echoreq max-pkt-rate 100/10
            pf.PFRule(action=pf.PF_PASS,
                      direction=pf.PF_IN,
                      af=AF_INET,
                      proto=IPPROTO_ICMP,
                      type=pf.ICMP_ECHO+1,
                      keep_state=pf.PF_STATE_NORMAL,
                      pktrate=pf.PFThreshold(100, 10))]

        rules[2].append(
            pf.PFTable("spammers", flags=pf.PFR_TFLAG_PERSIST),
            # pass in on $ifname inet proto tcp from ! <spammers> \
            #    to $ifname port 25 rdr-to 10.0.1.23
            pf.PFRule(action=pf.PF_PASS,
                      direction=pf.PF_IN,
                      ifname=self.testif,
                      af=AF_INET,
                      proto=IPPROTO_TCP,
                      src=pf.PFRuleAddr(pf.PFAddr("<spammers>"), neg=True),
                      dst=pf.PFRuleAddr(iface, pf.PFPort(25, IPPROTO_TCP)),
                      flags="S", flagset="SA",
                      keep_state=pf.PF_STATE_NORMAL,
                      rdr=pf.PFPool(pf.PF_POOL_RDR, pf.PFAddr("10.0.1.23"))))
            
        self.pf.clear_rules()
        rs = pf.PFRuleset()
        rs.append(*tables)
        rs.append(*rules)
        self.pf.load_ruleset(rs)
        self.assertEqual(len(self.pf.get_ruleset().rules), len(rules))

    def test_set_syncookies(self):
        modes = (pf.PF_SYNCOOKIES_NEVER, "always")
        # As far as I know, there's no way to retrieve the current syncookies
        # mode from the system.
        # So I assume that if no exception was raised, everything is OK.
        for mode in modes:
            self.pf.set_syncookies(mode)

    def test_set_synflood_watermarks(self):
        hiwat, lowat = 2000, 1000
        _hiwat, _lowat = self.pf.get_synflood_watermarks()
        self.pf.set_synflood_watermarks(hiwat, lowat)
        self.assertEqual(self.pf.get_synflood_watermarks(), (hiwat, lowat))
        self.pf.set_synflood_watermarks(_hiwat, _lowat)

    def test_add_tables(self):
        tblname = "test_table"
        self.assertEqual(self._add_table(tblname), 1)
        self.assertEqual(self.pf.get_tables()[0].name, tblname)

    def test_clear_tables(self):
        self.assertEqual(self._add_table(), 1)
        self.pf.clear_tables()
        self.assertFalse(self.pf.get_tables())

    def test_del_tables(self):
        self.assertEqual(self._add_table(), 1)
        table = self.pf.get_tables()[0]
        self.pf.del_tables(table)
        self.assertFalse(self.pf.get_tables())

    def test_test_addrs(self):
        self.assertEqual(self._add_table(), 1)
        table = self.pf.get_tables()[0]
        self.assertTrue(self.pf.test_addrs(table, "10.0.1.10"))
        self.assertFalse(self.pf.test_addrs(table, "10.0.1.20"))

    def test_add_addrs(self):
        self.assertEqual(self._add_table(), 1)
        table = self.pf.get_tables()[0]
        addrs = ["10.0.1.11", "10.0.1.12"]
        self.pf.clear_addrs(table)
        self.assertEqual(self.pf.add_addrs(table, *addrs), len(addrs))
        self.assertEqual(len(self.pf.get_addrs(table)), len(addrs))

    def test_clear_addrs(self):
        self.assertEqual(self._add_table(), 1)
        table = self.pf.get_tables()[0]
        self.assertTrue(self.pf.clear_addrs(table))
        self.assertFalse(self.pf.get_addrs(table))

    def test_del_addrs(self):
        self.assertEqual(self._add_table(), 1)
        table = self.pf.get_tables()[0]
        addrs = ["10.0.1.11", "10.0.1.12"]
        self.assertEqual(self.pf.set_addrs(table, *addrs)[1], len(addrs))
        self.assertEqual(self.pf.del_addrs(table, addrs.pop()), 1)
        self.assertEqual(len(self.pf.get_addrs(table)), len(addrs))

    def test_set_addrs(self):
        self.assertEqual(self._add_table(), 1)
        table = self.pf.get_tables()[0]
        addrs = ["10.0.1.11", "10.0.1.12"]
        self.assertEqual(self.pf.set_addrs(table, *addrs)[1], len(addrs))
        self.assertEqual(len(self.pf.get_addrs(table)), len(addrs))

    def test_get_tstats(self):
        self.assertEqual(self._add_table(), 1)
        self.assertTrue(self.pf.get_tstats())

    def test_clear_tstats(self):
        self.assertEqual(self._add_table(), 1)
        table = self.pf.get_tables()[0]
        self.assertEqual(self.pf.clear_tstats(table), 1)

    def _add_table(self, tblname="test_table"):
        table = pf.PFTable(tblname, "10.0.1.10",
                           flags=pf.PFR_TFLAG_PERSIST)
        self.pf.clear_tables()
        return self.pf.add_tables(table)

    def _create_state(self):
        ruleset = pf.PFRuleset()
        ruleset.append(pf.PFRule(action=pf.PF_PASS,
                                 flags="S", flagset="SA",
                                 proto=IPPROTO_UDP,
                                 keep_state=pf.PF_STATE_NORMAL))
        self.pf.load_ruleset(ruleset)
        self.pf.clear_states()
        with open(os.devnull, "w") as n:
            subprocess.call(["/usr/bin/nslookup", "google.com"],
                            stdout=n, stderr=n)



"""Export constants shared by all classes of the module."""


# Actions (from /usr/include/net/pfvar.h)
PF_PASS                 = 0
PF_DROP                 = 1
PF_SCRUB                = 2
PF_NOSCRUB              = 3
PF_NAT                  = 4
PF_NONAT                = 5
PF_BINAT                = 6
PF_NOBINAT              = 7
PF_RDR                  = 8
PF_NORDR                = 9
PF_SYNPROXY_DROP        = 10
PF_DEFER                = 11
PF_MATCH                = 12
PF_DIVERT               = 13
PF_RT                   = 14
PF_AFRT                 = 15

# PF transaction types (from /usr/include/net/pfvar.h)
PF_TRANS_RULESET        = 0
PF_TRANS_TABLE          = 1

# PF rule flags (from /usr/include/net/pfvar.h)
PFRULE_DROP             = 0x000000
PFRULE_RETURNRST        = 0x000001
PFRULE_FRAGMENT         = 0x000002
PFRULE_RETURNICMP       = 0x000004
PFRULE_RETURN           = 0x000008
PFRULE_NOSYNC           = 0x000010
PFRULE_SRCTRACK         = 0x000020
PFRULE_RULESRCTRACK     = 0x000040
PFRULE_SETDELAY         = 0x000080
PFRULE_IFBOUND          = 0x010000
PFRULE_STATESLOPPY      = 0x020000
PFRULE_PFLOW            = 0x040000
PFRULE_ONCE             = 0x100000
PFRULE_AFTO             = 0x200000
PFRULE_EXPIRED          = 0x400000

# PF rule flags (from /usr/include/net/pfvar.h)
PFRULE_IFBOUND          = 0x00010000
PFRULE_STATESLOPPY      = 0x00020000
PFRULE_PFLOW            = 0x00040000
PFRULE_ONCE             = 0x00100000
PFRULE_AFTO             = 0x00200000

# Port comparison operators (from /usr/include/net/pfvar.h)
PF_OP_NONE              = 0
PF_OP_IRG               = 1
PF_OP_EQ                = 2
PF_OP_NE                = 3
PF_OP_LT                = 4
PF_OP_LE                = 5
PF_OP_GT                = 6
PF_OP_GE                = 7
PF_OP_XRG               = 8
PF_OP_RRG               = 9

# Rules retrieval options (from /usr/include/net/pfvar.h)
PF_GET_NONE             = 0
PF_GET_CLR_CNTR         = 1

# PF keep states (from /usr/include/net/pfvar.h)
PF_STATE_NORMAL         = 0x1
PF_STATE_MODULATE       = 0x2
PF_STATE_SYNPROXY       = 0x3

# Routing options (from /usr/include/net/pfvar.h)
PF_NOPFROUTE            = 0
PF_ROUTETO              = 1
PF_DUPTO                = 2
PF_REPLYTO              = 3

# State keys (from /usr/include/net/pfvar.h)
PF_SK_WIRE              = 0
PF_SK_STACK             = 1
PF_SK_BOTH              = 2

# Log options (from /usr/include/net/pfvar.h)
PF_LOG                  = 0x01
PF_LOG_ALL              = 0x02
PF_LOG_SOCKET_LOOKUP    = 0x04
PF_LOG_FORCE            = 0x08
PF_LOG_MATCHES          = 0x10

# Address types (from /usr/include/net/pfvar.h)
PF_ADDR_ADDRMASK        = 0
PF_ADDR_NOROUTE         = 1
PF_ADDR_DYNIFTL         = 2
PF_ADDR_TABLE           = 3
PF_ADDR_RTLABEL         = 4
PF_ADDR_URPFFAILED      = 5
PF_ADDR_RANGE           = 6
PF_ADDR_NONE            = 7

# OS fingerprints matches (from /usr/include/net/pfvar.h)
PF_OSFP_ANY             = 0
PF_OSFP_UNKNOWN         = -1
PF_OSFP_NOMATCH         = -2

# Interface flags (from /usr/include/net/pfvar.h)
PFI_AFLAG_NETWORK       = 0x01
PFI_AFLAG_BROADCAST     = 0x02
PFI_AFLAG_PEER          = 0x04
PFI_AFLAG_MODEMASK      = 0x07
PFI_AFLAG_NOALIAS       = 0x08

# Traffic directions (from /usr/include/net/pfvar.h)
PF_INOUT                = 0
PF_IN                   = 1
PF_OUT                  = 2
PF_FWD                  = 3

# Flush options (from /usr/include/net/pfvar.h)
PF_FLUSH                = 0x01
PF_FLUSH_GLOBAL         = 0x02

# IP type of service (from /usr/include/netinet/ip.h)
IPTOS_LOWDELAY          = 0x10
IPTOS_THROUGHPUT        = 0x08
IPTOS_RELIABILITY       = 0x04

# NAT ports range (from /usr/src/sbin/pfctl/pfctl_parser.h)
PF_NAT_PROXY_PORT_LOW   = 50001
PF_NAT_PROXY_PORT_HIGH  = 65535

# Thresholds for syncookies (from /usr/include/net/pfvar.h)
PF_THRESHOLD_MULT       = 1000
PF_THRESHOLD_MAX        = 0xffffffff / PF_THRESHOLD_MULT

# Divert types (from /usr/include/net/pfvar.h)
PF_DIVERT_NONE          = 0
PF_DIVERT_TO            = 1
PF_DIVERT_REPLY         = 2
PF_DIVERT_PACKET        = 3

# Syncookies (from /usr/include/net/pfvar.h)
PF_SYNCOOKIES_NEVER	= 0
PF_SYNCOOKIES_ALWAYS	= 1
PF_SYNCOOKIES_ADAPTIVE	= 2
PF_SYNCOOKIES_MODE_MAX	= PF_SYNCOOKIES_ADAPTIVE

# Pool IDs (from /usr/src/sbin/pfctl/pfctl_parser.c)
PF_POOL_ROUTE           = 0
PF_POOL_NAT             = 1
PF_POOL_RDR             = 2

# Pool options (from /usr/include/net/pfvar.h)
PF_POOL_TYPEMASK        = 0x0f
PF_POOL_STICKYADDR      = 0x20

# Pool types (from /usr/include/net/pfvar.h)
PF_POOL_NONE            = 0
PF_POOL_BITMASK         = 1
PF_POOL_RANDOM          = 2
PF_POOL_SRCHASH         = 3
PF_POOL_ROUNDROBIN      = 4
PF_POOL_LEASTSTATES     = 5

# Mask for window scaling factor (from /usr/include/net/pfvar.h)
PF_WSCALE_MASK          = 0x0f

# Debug levels (from /usr/include/sys/syslog.h)
LOG_EMERG               = 0
LOG_ALERT               = 1
LOG_CRIT                = 2
LOG_ERR                 = 3
LOG_WARNING             = 4
LOG_NOTICE              = 5
LOG_INFO                = 6
LOG_DEBUG               = 7

# The 'unlimited' value for limits on the memory pools
UINT_MAX                = 0xffffffff

# Limits (from /usr/include/net/pfvar.h)
PF_LIMIT_STATES         = 0
PF_LIMIT_SRC_NODES      = 1
PF_LIMIT_FRAGS          = 2
PF_LIMIT_TABLES         = 3
PF_LIMIT_TABLE_ENTRIES  = 4
PF_LIMIT_PKTDELAY_PKTS  = 5
PF_LIMIT_MAX            = 6

# Timeouts (from /usr/include/net/pfvar.h)
PFTM_TCP_FIRST_PACKET   = 0
PFTM_TCP_OPENING        = 1
PFTM_TCP_ESTABLISHED    = 2
PFTM_TCP_CLOSING        = 3
PFTM_TCP_FIN_WAIT       = 4
PFTM_TCP_CLOSED         = 5
PFTM_UDP_FIRST_PACKET   = 6
PFTM_UDP_SINGLE         = 7
PFTM_UDP_MULTIPLE       = 8
PFTM_ICMP_FIRST_PACKET  = 9
PFTM_ICMP_ERROR_REPLY   = 10
PFTM_OTHER_FIRST_PACKET = 11
PFTM_OTHER_SINGLE       = 12
PFTM_OTHER_MULTIPLE     = 13
PFTM_FRAG               = 14
PFTM_INTERVAL           = 15
PFTM_ADAPTIVE_START     = 16
PFTM_ADAPTIVE_END       = 17
PFTM_SRC_NODE           = 18
PFTM_TS_DIFF            = 19
PFTM_MAX                = 20
PFTM_PURGE              = 21
PFTM_UNLINKED           = 22

# TCP States (from /usr/include/netinet/tcp_fsm.h)
TCPS_CLOSED             = 0
TCPS_LISTEN             = 1
TCPS_SYN_SENT           = 2
TCPS_SYN_RECEIVED       = 3
TCPS_ESTABLISHED        = 4
TCPS_CLOSE_WAIT         = 5
TCPS_FIN_WAIT_1         = 6
TCPS_CLOSING            = 7
TCPS_LAST_ACK           = 8
TCPS_FIN_WAIT_2         = 9
TCPS_TIME_WAIT          = 10
TCP_NSTATES             = 11

# From /usr/include/net/pfvar.h
PF_TCPS_PROXY_SRC       = TCP_NSTATES + 0
PF_TCPS_PROXY_DST       = TCP_NSTATES + 1

# UDP state enumeration (from /usr/include/net/pfvar.h)
PFUDPS_NO_TRAFFIC       = 0
PFUDPS_SINGLE           = 1
PFUDPS_MULTIPLE         = 2
PFUDPS_NSTATES          = 3

# States for non-TCP protocols (from /usr/include/net/pfvar.h)
PFOTHERS_NO_TRAFFIC     = 0
PFOTHERS_SINGLE         = 1
PFOTHERS_MULTIPLE       = 2
PFOTHERS_NSTATES        = 3

# Pfsync flags (from /usr/include/net/pfvar.h)
PFSYNC_FLAG_SRCNODE     = 0x04
PFSYNC_FLAG_NATSRCNODE  = 0x08

# PF states flags (from /usr/include/net/pfvar.h)
PFSTATE_ALLOWOPTS       = 0x0001
PFSTATE_SLOPPY          = 0x0002
PFSTATE_PFLOW           = 0x0004
PFSTATE_NOSYNC          = 0x0008
PFSTATE_ACK             = 0x0010
PFSTATE_NODF            = 0x0020
PFSTATE_SETTOS          = 0x0040
PFSTATE_RANDOMID        = 0x0080
PFSTATE_SCRUB_TCP       = 0x0100
PFSTATE_SETPRIO         = 0x0200
PFSTATE_SCRUBMASK       = PFSTATE_NODF|PFSTATE_RANDOMID|PFSTATE_SCRUB_TCP
PFSTATE_SETMASK         = PFSTATE_SETTOS|PFSTATE_SETPRIO

# Reassembly flags (from /usr/include/net/pfvar.h)
PF_REASS_ENABLED        = 0x01
PF_REASS_NODF           = 0x02

# Table flags (from /usr/include/net/pfvar.h)
PFR_TFLAG_PERSIST       = 0x01
PFR_TFLAG_CONST         = 0x02
PFR_TFLAG_ACTIVE        = 0x04
PFR_TFLAG_INACTIVE      = 0x08
PFR_TFLAG_REFERENCED    = 0x10
PFR_TFLAG_REFDANCHOR    = 0x20
PFR_TFLAG_COUNTERS      = 0x40
PFR_TFLAG_USRMASK       = 0x43
PFR_TFLAG_SETMASK       = 0x3C
PFR_TFLAG_ALLMASK       = 0x7F

PFR_FLAG_DUMMY          = 0x00000002
PFR_FLAG_FEEDBACK       = 0x00000004
PFR_FLAG_CLSTATS        = 0x00000008
PFR_FLAG_ADDRSTOO       = 0x00000010
PFR_FLAG_REPLACE        = 0x00000020
PFR_FLAG_ALLRSETS       = 0x00000040
PFR_FLAG_ALLMASK        = 0x0000007f

PFR_DIR_IN              = 0
PFR_DIR_OUT             = 1
PFR_DIR_MAX             = 2

PFR_OP_BLOCK            = 0
PFR_OP_MATCH            = 1
PFR_OP_PASS             = 2
PFR_OP_ADDR_MAX         = 3
PFR_OP_TABLE_MAX        = 4
PFR_OP_XPASS            = PFR_OP_ADDR_MAX

PFR_REFCNT_RULE         = 0
PFR_REFCNT_ANCHOR       = 1
PFR_REFCNT_MAX          = 2

# pfrke type (from /usr/include/net/pfvar.h)
PFRKE_PLAIN             = 0
PFRKE_ROUTE             = 1
PFRKE_COST              = 2
PFRKE_MAX               = 3

# Interface flags (from /usr/include/net/pfvar.h)
PFI_IFLAG_SKIP          = 0x0100
PFI_IFLAG_ANY		    = 0x0200

# From /usr/src/sbin/pfctl/pfctl.h
DEFAULT_PRIORITY        = 1
DEFAULT_QLIMIT          = 50

# Queue flags (from /usr/include/net/pfvar.h)
PFQS_FLOWQUEUE          = 0x0001
PFQS_ROOTCLASS          = 0x0002
PFQS_DEFAULT            = 0x1000

# Match "prio 0" packets
PF_PRIO_ZERO            = 0xff

# ICMP types (from /usr/include/netinet/ip_icmp.h)
ICMP_ECHO                        = 8
ICMP_ECHOREPLY                   = 0
ICMP_UNREACH                     = 3
ICMP_SOURCEQUENCH                = 4
ICMP_REDIRECT                    = 5
ICMP_ALTHOSTADDR                 = 6
ICMP_ROUTERADVERT                = 9
ICMP_ROUTERSOLICIT               = 10
ICMP_TIMXCEED                    = 11
ICMP_PARAMPROB                   = 12
ICMP_TSTAMP                      = 13
ICMP_TSTAMPREPLY                 = 14
ICMP_IREQ                        = 15
ICMP_IREQREPLY                   = 16
ICMP_MASKREQ                     = 17
ICMP_MASKREPLY                   = 18
ICMP_TRACEROUTE                  = 30
ICMP_DATACONVERR                 = 31
ICMP_MOBILE_REDIRECT             = 32
ICMP_IPV6_WHEREAREYOU            = 33
ICMP_IPV6_IAMHERE                = 34
ICMP_MOBILE_REGREQUEST           = 35
ICMP_MOBILE_REGREPLY             = 36
ICMP_SKIP                        = 39
ICMP_PHOTURIS                    = 40

# ICMP codes (from /usr/include/netinet/ip_icmp.h)
ICMP_UNREACH_NET                 = 0    # Destination unreachable
ICMP_UNREACH_HOST                = 1
ICMP_UNREACH_PROTOCOL            = 2
ICMP_UNREACH_PORT                = 3
ICMP_UNREACH_NEEDFRAG            = 4
ICMP_UNREACH_SRCFAIL             = 5
ICMP_UNREACH_NET_UNKNOWN         = 6
ICMP_UNREACH_HOST_UNKNOWN        = 7
ICMP_UNREACH_ISOLATED            = 8
ICMP_UNREACH_NET_PROHIB          = 9
ICMP_UNREACH_HOST_PROHIB         = 10
ICMP_UNREACH_TOSNET              = 11
ICMP_UNREACH_TOSHOST             = 12
ICMP_UNREACH_FILTER_PROHIB       = 13
ICMP_UNREACH_HOST_PRECEDENCE     = 14
ICMP_UNREACH_PRECEDENCE_CUTOFF   = 15
ICMP_REDIRECT_NET                = 0    # Shorter route
ICMP_REDIRECT_HOST               = 1
ICMP_REDIRECT_TOSNET             = 2
ICMP_REDIRECT_TOSHOST            = 3
ICMP_ROUTERADVERT_NORMAL         = 0    # Router advertisement
ICMP_ROUTERADVERT_NOROUTE_COMMON = 16
ICMP_TIMXCEED_INTRANS            = 0    # Time exceeded
ICMP_TIMXCEED_REASS              = 1
ICMP_PARAMPROB_ERRATPTR          = 0    # IP header bad
ICMP_PARAMPROB_OPTABSENT         = 1
ICMP_PARAMPROB_LENGTH            = 2
ICMP_PHOTURIS_UNKNOWN_INDEX      = 1    # Photuris
ICMP_PHOTURIS_AUTH_FAILED        = 2
ICMP_PHOTURIS_DECRYPT_FAILED     = 3

# ICMP6 types (from /usr/include/netinet/icmp6.h)
ICMP6_DST_UNREACH                = 1
ICMP6_PACKET_TOO_BIG             = 2
ICMP6_TIME_EXCEEDED              = 3
ICMP6_PARAM_PROB                 = 4
ICMP6_ECHO_REQUEST               = 128
ICMP6_ECHO_REPLY                 = 129
ICMP6_MEMBERSHIP_QUERY           = 130
MLD_LISTENER_QUERY               = 130
ICMP6_MEMBERSHIP_REPORT          = 131
MLD_LISTENER_REPORT              = 131
ICMP6_MEMBERSHIP_REDUCTION       = 132
MLD_LISTENER_DONE                = 132
ND_ROUTER_SOLICIT                = 133
ND_ROUTER_ADVERT                 = 134
ND_NEIGHBOR_SOLICIT              = 135
ND_NEIGHBOR_ADVERT               = 136
ND_REDIRECT                      = 137
ICMP6_ROUTER_RENUMBERING         = 138
ICMP6_WRUREQUEST                 = 139
ICMP6_WRUREPLY                   = 140
ICMP6_FQDN_QUERY                 = 139
ICMP6_FQDN_REPLY                 = 140
ICMP6_NI_QUERY                   = 139
ICMP6_NI_REPLY                   = 140
MLD_MTRACE_RESP                  = 200
MLD_MTRACE                       = 201

# ICMP6 codes (from /usr/include/netinet/icmp6.h)
ICMP6_DST_UNREACH_NOROUTE        = 0
ICMP6_DST_UNREACH_ADMIN          = 1
ICMP6_DST_UNREACH_NOTNEIGHBOR    = 2
ICMP6_DST_UNREACH_BEYONDSCOPE    = 2
ICMP6_DST_UNREACH_ADDR           = 3
ICMP6_DST_UNREACH_NOPORT         = 4
ICMP6_TIME_EXCEED_TRANSIT        = 0
ICMP6_TIME_EXCEED_REASSEMBLY     = 1
ICMP6_PARAMPROB_HEADER           = 0
ICMP6_PARAMPROB_NEXTHEADER       = 1
ICMP6_PARAMPROB_OPTION           = 2
ND_REDIRECT_ONLINK               = 0
ND_REDIRECT_ROUTER               = 1



"""A package for managing OpenBSD's Packet Filter."""


__copyright__ = """
Copyright (c) 2008-2022, Daniele Mazzocchio
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright notice, this
     list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.
   * Neither the name of the developer nor the names of its contributors may be
     used to endorse or promote products derived from this software without
     specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import os


__author__  = "Daniele Mazzocchio <danix@kernel-panic.it>"
__version__ = "0.2.2"

__OBSD_VERSION__ = "7.0"


from pf.exceptions import PFError
from pf.constants import *
from pf.status import *
from pf.state import *
from pf.table import *
from pf.rule import *
from pf.queue import *
from pf.filter import *

import pf.lib


__all__ = ['PFError',
           'PFStatus',
           'PFIface',
           'PFUid',
           'PFGid',
           'PFAddr',
           'PFPort',
           'PFRuleAddr',
           'PFPool',
           'PFRule',
           'PFRuleset',
           'PFStatePeer',
           'PFStateKey',
           'PFState',
           'PFTableAddr',
           'PFTable',
           'PFTStats',
           'ServiceCurve',
           'PFQueue',
           'PacketFilter']

import pf.constants
__all__.extend(os._get_exports_list(pf.constants))
del pf.constants



"""Classes representing the entries in the firewall's state table."""

from socket import *
from ctypes import *
from struct import *

from pf.constants import *
from pf._struct import *
from pf._base import PFObject
from pf._utils import getprotobynumber, tcpstates, udpstates
from pf.rule import PFAddr, PFPort


__all__ = ['PFStatePeer',
           'PFStateKey',
           'PFState']


class PFStatePeer(PFObject):
    """Represents a connection endpoint."""

    _struct_type = pfsync_state_peer

    def __init__(self, peer):
        """Check argument and initialize class attributes."""
        super(PFStatePeer, self).__init__(peer)

    def _from_struct(self, p):
        """Initialize class attributes from a pfsync_state_peer structure."""
        self.seqlo       = ntohl(p.seqlo)
        self.seqhi       = ntohl(p.seqhi)
        self.seqdiff     = ntohl(p.seqdiff)
        self.max_win     = p.max_win
        self.mss         = p.mss
        self.state       = p.state
        self.wscale      = p.wscale

        self.pfss_flags  = p.scrub.pfss_flags
        self.pfss_ttl    = p.scrub.pfss_ttl
        self.scrub_flag  = p.scrub.scrub_flag
        self.pfss_ts_mod = p.scrub.pfss_ts_mod


class PFStateKey(PFObject):
    """Represents a state key."""

    _struct_type = pfsync_state_key

    def __init__(self, key, af):
        """Check argument and initialize class attributes."""
        self.af = af
        super(PFStateKey, self).__init__(key)

    def _from_struct(self, k):
        """Initialize class attributes from a pfsync_state_key structure."""
        a = (pf_addr_wrap(), pf_addr_wrap())


        a[0].v.a.addr, a[1].v.a.addr = k.addr
        mask = b'\xff' * {AF_INET: 4, AF_INET6: 16}[self.af]
        memmove(a[0].v.a.mask.v6, c_char_p(mask), len(mask))
        memmove(a[1].v.a.mask.v6, c_char_p(mask), len(mask))

        self.addr    = (PFAddr(a[0], self.af), PFAddr(a[1], self.af))
        self.port    = (PFPort(ntohs(k.port[0])), PFPort(ntohs(k.port[1])))
        self.rdomain = ntohs(k.rdomain)


class PFState(PFObject):
    """Represents an entry in Packet Filter's state table."""

    _struct_type = pfsync_state

    def __init__(self, state):
        """Check argument and initialize class attributes."""
        super(PFState, self).__init__(state)

    def _from_struct(self, s):
        """Initialize class attributes from a pfsync_state structure."""
        self.id              = unpack("Q", pack(">Q", s.id))[0]
        self.ifname          = s.ifname.decode()

        a                    = pf_addr_wrap()
        a.v.a.addr           = s.rt_addr
        self.rt_addr         = PFAddr(a, s.af)

        self.rule            = ntohl(s.rule)
        self.anchor          = ntohl(s.anchor)
        self.nat_rule        = ntohl(s.nat_rule)
        self.creation        = ntohl(s.creation)
        self.expire          = ntohl(s.expire)

        p = unpack('>IIII', string_at(addressof(s.packets), sizeof(s.packets)))
        self.packets         = ((p[0] << 32 | p[1]), (p[2] << 32 | p[3]))
        b = unpack('>IIII', string_at(addressof(s.bytes), sizeof(s.bytes)))
        self.bytes           = ((b[0] << 32 | b[1]), (b[2] << 32 | b[3]))

        self.creatorid       = ntohl(s.creatorid) & 0xffffffff
        self.rtableid        = s.rtableid
        self.max_mss         = s.max_mss
        self.af              = s.af
        self.proto           = s.proto
        self.direction       = s.direction
        self.log             = s.log
        self.timeout         = s.timeout
        self.sync_flags      = s.sync_flags
        self.updates         = s.updates
        self.min_ttl         = s.min_ttl
        self.set_tos         = s.set_tos
        self.state_flags     = s.state_flags

        if self.direction == PF_OUT:
            self.src         = PFStatePeer(s.src)
            self.dst         = PFStatePeer(s.dst)
            self.sk          = PFStateKey(s.key[PF_SK_STACK], s.af)
            self.nk          = PFStateKey(s.key[PF_SK_WIRE], s.af)
            if self.proto in (IPPROTO_ICMP, IPPROTO_ICMPV6):
                self.sk.port = (self.nk.port[0], self.sk.port[1])
        else:
            self.src         = PFStatePeer(s.dst)
            self.dst         = PFStatePeer(s.src)
            self.sk          = PFStateKey(s.key[PF_SK_WIRE], s.af)
            self.nk          = PFStateKey(s.key[PF_SK_STACK], s.af)
            if self.proto in (IPPROTO_ICMP, IPPROTO_ICMPV6):
                self.sk.port = (self.sk.port[0], self.nk.port[1])

    def _to_string(self):
        """Return a string representing the state."""
        sk, nk   = self.sk, self.nk
        sp       = (sk.port[0].num[0], sk.port[1].num[0])
        np       = (nk.port[0].num[0], nk.port[1].num[0])
        src, dst = self.src, self.dst
        afto     = (nk.af != sk.af)

        s  = "{.ifname} ".format(self)
        s += "{} ".format(getprotobynumber(self.proto) or self.proto)

        s += "{0.addr[1]}:{0.port[1]}".format(nk)
        if afto or (nk.addr[1] != sk.addr[1]) or (np[1] != sp[1]) or \
           (nk.rdomain != sk.rdomain):
            i = int(not afto)
            s += " ({}:{})".format(sk.addr[i], sk.port[i])

        if self.direction == PF_OUT or (afto and self.direction == PF_IN):
            s += " -> "
        else:
            s += " <- "

        s += "{0.addr[0]}:{0.port[0]}".format(nk)
        if afto or (nk.addr[1] != nk.addr[1]) or (np[1] != sp[1]) or \
           (nk.rdomain != sk.rdomain):
            i = int(afto)
            s += " ({}:{})".format(sk.addr[i], sk.port[i])

        s += "       "
        if self.proto == IPPROTO_TCP:
            if (src.state <= TCPS_TIME_WAIT and
                dst.state <= TCPS_TIME_WAIT):
                s += "{}:{}".format(tcpstates[src.state],
                                    tcpstates[dst.state])
            elif (src.state == PF_TCPS_PROXY_SRC or
                  dst.state == PF_TCPS_PROXY_SRC):
                s += "PROXY:SRC"
            elif (src.state == PF_TCPS_PROXY_DST or
                  dst.state == PF_TCPS_PROXY_DST):
                s += "PROXY:DST"
            else:
                s += "<BAD STATE LEVELS {.state}:{.state}>".format(src, dst)

            s += "\n"
            for p in src, dst:
                s += "   [{.seqlo} + {}]".format(p, (p.seqhi - p.seqlo))
                if p.seqdiff:
                    s += "(+{.seqdiff})".format(p)
                if src.wscale and dst.wscale:
                    s += " wscale {}".format(p.wscale & PF_WSCALE_MASK)

        elif (self.proto == IPPROTO_UDP  and
              src.state < PFUDPS_NSTATES and dst.state < PFUDPS_NSTATES) or \
             (self.proto not in (IPPROTO_ICMP, IPPROTO_ICMPV6) and
              src.state < PFOTHERS_NSTATES and dst.state < PFOTHERS_NSTATES):
            s += "{}:{}".format(udpstates[src.state], udpstates[dst.state])
        else:
            s += "{.state}:{.state}".format(src, dst)

        hrs, sec = divmod(self.creation, 60)
        hrs, min = divmod(hrs, 60)
        s += "\n   age {:02d}:{:02d}:{:02d}".format(hrs, min, sec)

        hrs, sec = divmod(self.expire, 60)
        hrs, min = divmod(hrs, 60)
        s += ", expires in {:02d}:{:02d}:{:02d}".format(hrs, min, sec)

        s += ", {0.packets[0]}:{0.packets[1]} pkts".format(self)
        s += ", {0.bytes[0]}:{0.bytes[1]} bytes".format(self)

        if self.anchor != 0xffffffff:
            s += ", anchor {0.anchor}" .format(self)
        if self.rule != 0xffffffff:
            s += ", rule {0.rule}".format(self)
        if self.state_flags & PFSTATE_SLOPPY:
            s += ", sloppy"
        if self.state_flags & PFSTATE_PFLOW:
            s += ", pflow"
        if self.sync_flags & PFSYNC_FLAG_SRCNODE:
            s += ", source-track"
        if self.sync_flags & PFSYNC_FLAG_NATSRCNODE:
            s += ", sticky-address\n"

        s += "\n   id: {0.id:016x} creatorid: {0.creatorid:08x}".format(self)
        if self.sync_flags & PFSTATE_NOSYNC:
            s += " (no-sync)"

        return s
